import os
import json
import re


class Autotrace:
    """A class to search for specific variables in files and save them in a
    JSON-compatible dictionary.
    """

    def __init__(self, search_path):
        """
        Initialize the Autotrace class with a search path.

        Args:
            search_path (str): The file path to search for variables.
        """
        self.search_path = search_path
        self.variables = {
            "dataset_name": None,
            "dataset_path": None,
            "prompt_name": None,
            "prompt_string": None,
            "model_name": None,
            "model_parameters": None,
        }
        self.variable_counter = 1

    def search_variables(self):
        """Search for the specified variables in the given file path and
        update the dictionary.
        """
        pattern = re.compile(r"(dataset_name|dataset_path|prompt_name|prompt_string|model_name|model_parameters)\s*=\s*['\"](.*?)['\"]")

        for root, _, files in os.walk(self.search_path):
            for file in files:
                if file.endswith(".py"):
                    file_path = os.path.join(root, file)
                    with open(file_path, 'r') as f:
                        content = f.read()
                        matches = pattern.findall(content)
                        for var, value in matches:
                            if self.variables[var] is None:
                                self.variables[var] = value

        for var in self.variables:
            if self.variables[var] is None:
                self.variables[var] = f"{var}_{self.variable_counter}"
                self.variable_counter += 1

    def check_json_compatibility(self):
        """Ensure the dictionary is JSON-compatible and transform if necessary.
        """
        try:
            json.dumps(self.variables)
        except TypeError:
            # Handle non-serializable values if necessary
            self.variables = {k: str(v) for k, v in self.variables.items()}

    def save_to_json(self, output_file):
        """Save the dictionary to a JSON file.

        Args:
            output_file (str): The path to the output JSON file.
        """
        self.check_json_compatibility()
        with open(output_file, 'w') as f:
            json.dump(self.variables, f, indent=4)

    def run(self, output_file):
        """Execute the variable search and save the result to a JSON file.

        Args:
            output_file (str): The path to the output JSON file.
        """
        self.search_variables()
        self.save_to_json(output_file)




"""Containerize your model for production easily."""
import os

class Containerize:
    """A class to evaluate model size and create an optimized Dockerfile for containerization.
    """

    BASE_IMAGES = {
        'small': 'python:3.9-slim',
        'medium': 'python:3.9',
        'large': 'ubuntu:20.04'
    }

    def __init__(self, model_path):
        """Initialize the Containerize class with the model path.

        Args:
            model_path (str): The file path to the model.
        """
        self.model_path = model_path
        self.model_size = self.evaluate_model_size()

    def evaluate_model_size(self):
        """Evaluate the size of the model.

        Returns:
            str: The size category of the model ('small', 'medium', 'large').
        """
        size_in_mb = os.path.getsize(self.model_path) / (1024 * 1024)
        if size_in_mb < 100:
            return 'small'
        elif size_in_mb < 500:
            return 'medium'
        else:
            return 'large'

    def select_base_image(self):
        """Select the best base image based on model size.

        Returns:
            str: The selected base image.
        """
        return self.BASE_IMAGES[self.model_size]

    def create_dockerfile(self, output_path):
        """Create a Dockerfile using a multi-stage build and optimize for production.

        Args:
            output_path (str): The file path to save the Dockerfile.
        """
        base_image = self.select_base_image()
        
        dockerfile_content = f"""
        # Stage 1: Build Stage
        FROM {base_image} AS builder

        # Install dependencies
        RUN apt-get update && apt-get install -y --no-install-recommends \\
            build-essential \\
            && rm -rf /var/lib/apt/lists/*

        # Set working directory
        WORKDIR /app

        # Copy model files
        COPY . .

        # Stage 2: Production Stage
        FROM {base_image}

        # Install necessary packages
        RUN apt-get update && apt-get install -y --no-install-recommends \\
            ca-certificates \\
            && rm -rf /var/lib/apt/lists/*

        # Set working directory
        WORKDIR /app

        # Copy from the build stage
        COPY --from=builder /app /app

        # Install Python packages
        RUN pip install --no-cache-dir -r requirements.txt

        # Set environment variables for production
        ENV PYTHONUNBUFFERED=1

        # Run the application
        CMD ["python", "your_model_script.py"]
        """
        
        with open(output_path, 'w') as file:
            file.write(dockerfile_content.strip())

    def run(self, output_path):
        """Execute the process to create an optimized Dockerfile.

        Args:
            output_path (str): The file path to save the Dockerfile.
        """
        self.create_dockerfile(output_path)




"""Pytorch bitnet application.
"""

import torch
import torch.nn as nn


class RMSNorm(nn.Module):
    """Implements Root Mean Square Normalization as proposed in the paper:
    'Root Mean Square Layer Normalization' by Zhang and Sennrich, 2019.

    Args:
        normalized_shape (int): Input shape from an expected input of size
            (batch_size, normalized_shape).
        eps (float, optional): A value added to the denominator for
        numerical stability. Default: 1e-8.
    """
    def __init__(self, normalized_shape, eps=1e-8):
        super(RMSNorm, self).__init__()
        self.normalized_shape = normalized_shape
        self.eps = eps
        self.weight = nn.Parameter(torch.ones(normalized_shape))

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for RMSNorm.

        Args:
            x (Tensor): Input tensor of shape (batch_size, normalized_shape).

        Returns:
            Tensor: Normalized tensor.
        """
        # Compute the root mean square of the input tensor
        rms = torch.sqrt(torch.mean(x**2, dim=-1, keepdim=True) + self.eps)

        # Normalize the input tensor
        x_normalized = x / rms

        # Scale the normalized tensor by the learnable weight
        return self.weight * x_normalized


def per_token_quantization(x: torch.Tensor) -> torch.Tensor:
    """Quantizes input tensor to 8 bits per token.

    Args:
        x (Tensor): Input tensor to be quantized.

    Returns:
        Tensor: Quantized tensor.
    """
    min_val = x.min()
    max_val = x.max()

    scale = (max_val - min_val) / 255.0
    zero_point = min_val

    q_x = ((x - zero_point) / scale).round().clamp(0, 255).to(torch.uint8)

    return q_x


class TorchBitNet(nn.Module):
    """Custom linear layer with RMSNorm and per-token 8-bit quantization.

    Args:
        input_dim (int): Dimension of the input features.
        output_dim (int): Dimension of the output features.
        norm_dim (int): Dimension for normalization.
        epsilon (float, optional): Small value to avoid division by zero in
        RMSNorm. Default is 1e-5.
    """
    def __init__(self, input_dim, output_dim, norm_dim, epsilon=1e-5):
        super(TorchBitNet, self).__init__()
        self.linear = nn.Linear(input_dim, output_dim)
        self.norm = RMSNorm(norm_dim, eps=epsilon)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass for the custom linear layer.

        Args:
            x (Tensor): Input tensor.

        Returns:
            Tensor: Quantized output tensor.
        """
        # Normalize the input
        x = self.norm(x)

        # Apply the linear transformation
        x = self.linear(x)

        # Quantize the output
        x = per_token_quantization(x)

        return x



"""Test iso."""
import unittest
import torch
from isoai.model import (
    RMSNorm, compute_frequency_tensor, reshape_freqs_for_broadcast,
    apply_rotary_embeddings, repeat_key_value_tensor, MultiheadAttention,
    FeedForward, TransformerBlock, GatingMechanism, MoETransformerBlock, MoETransformer
)

class TestModel(unittest.TestCase):
    
    def setUp(self):
        # Setting up common test data
        self.args = type('', (), {})()  # Mock object for arguments
        self.args.vocab_size = 30522
        self.args.dim = 768
        self.args.n_heads = 12
        self.args.n_kv_heads = 12
        self.args.max_seq_len = 512
        self.args.norm_eps = 1e-5
        self.args.multiple_of = 64
        self.args.ffn_dim_multiplier = 4
        self.args.max_batch_size = 32
        self.args.num_experts = 4
        
        self.tokens = torch.randint(0, self.args.vocab_size, (2, self.args.max_seq_len)).cuda()
        self.freqs_cis = compute_frequency_tensor(self.args.dim // self.args.n_heads, self.args.max_seq_len * 2)

    def test_rmsnorm(self):
        rmsnorm = RMSNorm(self.args.dim)
        x = torch.randn(2, self.args.max_seq_len, self.args.dim).cuda()
        output = rmsnorm(x)
        self.assertEqual(output.shape, x.shape)

    def test_compute_frequency_tensor(self):
        freqs_cis = compute_frequency_tensor(8, 10)
        self.assertEqual(freqs_cis.shape, (10, 4))

    def test_reshape_freqs_for_broadcast(self):
        x = torch.randn(4, 10, 16)
        freqs_complex = torch.randn(10, 8, dtype=torch.complex64)
        reshaped_freqs = reshape_freqs_for_broadcast(freqs_complex, x)
        self.assertEqual(reshaped_freqs.shape, (1, 10, 1, 8))

    def test_apply_rotary_embeddings(self):
        xq = torch.randn(4, 10, 16)
        xk = torch.randn(4, 10, 16)
        xq_rotary, xk_rotary = apply_rotary_embeddings(xq, xk, self.freqs_cis)
        self.assertEqual(xq_rotary.shape, xq.shape)
        self.assertEqual(xk_rotary.shape, xk.shape)

    def test_repeat_key_value_tensor(self):
        x = torch.randn(4, 10, 8, 16)
        repeated_tensor = repeat_key_value_tensor(x, 2)
        self.assertEqual(repeated_tensor.shape, (4, 10, 16, 16))

    def test_multihead_attention(self):
        multihead_attention = MultiheadAttention(self.args)
        x = torch.randn(2, 10, self.args.dim).cuda()
        output = multihead_attention(x, start_pos=0, freqs_cis=self.freqs_cis)
        self.assertEqual(output.shape, x.shape)

    def test_feed_forward(self):
        feed_forward = FeedForward(
            dim=self.args.dim,
            hidden_dim=4 * self.args.dim,
            multiple_of=self.args.multiple_of,
            ffn_dim_multiplier=self.args.ffn_dim_multiplier
        )
        x = torch.randn(2, self.args.dim).cuda()
        output = feed_forward(x)
        self.assertEqual(output.shape, x.shape)

    def test_transformer_block(self):
        transformer_block = TransformerBlock(layer_id=0, args=self.args)
        x = torch.randn(2, 10, self.args.dim).cuda()
        output = transformer_block(x, start_pos=0, freqs_cis=self.freqs_cis)
        self.assertEqual(output.shape, x.shape)

    def test_gating_mechanism(self):
        gating = GatingMechanism(input_dim=self.args.dim, num_experts=self.args.num_experts)
        x = torch.randn(2, self.args.dim).cuda()
        gate_values = gating(x)
        self.assertEqual(gate_values.shape, (2, self.args.num_experts))

    def test_moe_transformer_block(self):
        moe_block = MoETransformerBlock(layer_id=0, args=self.args)
        x = torch.randn(2, 10, self.args.dim).cuda()
        output = moe_block(x, start_pos=0, freqs_cis=self.freqs_cis)
        self.assertEqual(output.shape, x.shape)

    def test_moe_transformer(self):
        moe_transformer = MoETransformer(self.args)
        output = moe_transformer(self.tokens, start_pos=0)
        self.assertEqual(output.shape, (2, self.args.max_seq_len, self.args.vocab_size))

if __name__ == '__main__':
    unittest.main()



"""Apply bitnet to LLAMA"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Optional, Tuple
from isobits.torch_bitnet import RMSNorm
from isobits.GPT_bitnet import compute_frequency_tensor
from isobits.GPT_bitnet import TransformerBlock

class BitNetLLAMA(nn.Module):
    """BitNet applied to a LLAMA model."""
    def __init__(self, params):
        super().__init__()
        self.tok_embeddings = nn.Embedding(params.vocab_size, params.dim)
        self.layers = nn.ModuleList([TransformerBlock(layer_id, params) for layer_id in range(params.n_layers)])
        self.norm = RMSNorm(params.dim, eps=params.norm_eps)
        self.output = nn.Linear(params.dim, params.vocab_size, bias=False)
        self.freqs_cis = compute_frequency_tensor(params.dim // params.n_heads, params.max_seq_len * 2)

    @torch.no_grad()
    def forward(self, tokens: torch.Tensor, start_pos: int) -> torch.Tensor:
        """Perform a forward pass through the BitNetLLAMA model.

        Args:
            tokens (torch.Tensor): Input token indices.
            start_pos (int): Starting position for attention caching.

        Returns:
            torch.Tensor: Output logits after applying the BitNetLLAMA model.
        """
        bsz, seqlen = tokens.shape
        h = self.tok_embeddings(tokens)
        self.freqs_cis = self.freqs_cis.to(h.device)
        freqs_cis = self.freqs_cis[start_pos : start_pos + seqlen]

        mask = None
        if seqlen > 1:
            mask = torch.full((seqlen, seqlen), float("-inf"), device=tokens.device)
            mask = torch.triu(mask, diagonal=1)
            mask = torch.hstack([torch.zeros((seqlen, start_pos), device=tokens.device), mask]).type_as(h)

        for layer in self.layers:
            h = layer(h, start_pos, freqs_cis, mask)
        h = self.norm(h)
        return self.output(h).float()




"""Quantize Huggingface models with BitNet."""

import torch.nn as nn
from isobits.torch_bitnet import TorchBitNet


def replace_linears_with_torchbitnet(model, norm_dim):
    """Replace all nn.Linear layers in the given model with TorchBitNet layers.

    Args:
        model (nn.Module): The PyTorch model in which to replace the layers.
        norm_dim (int): Dimension for normalization in TorchBitNet.
    """
    for name, module in model.named_children():
        if isinstance(module, nn.Linear):
            in_features = module.in_features
            out_features = module.out_features

            setattr(
                model,
                name,
                TorchBitNet(
                    input_dim=in_features,
                    output_dim=out_features,
                    norm_dim=norm_dim,
                )
            )
        else:
            # Recursively apply to child modules
            replace_linears_with_torchbitnet(module, norm_dim)



"""Apply bitnet to MoE"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from typing import Optional


class GatingMechanism(nn.Module):
    """Gating mechanism to route inputs to experts."""
    def __init__(self, input_dim: int, num_experts: int):
        super().__init__()
        self.num_experts = num_experts
        self.gate = nn.Linear(input_dim, num_experts)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """Forward pass of the gating mechanism."""
        logits = self.gate(x)
        gate_values = F.softmax(logits, dim=-1)
        return gate_values

class MoETransformerBlock(nn.Module):
    """Mixture of Experts Transformer Block."""
    def __init__(self, layer_id: int, args):
        super().__init__()
        self.num_experts = args.num_experts
        self.experts = nn.ModuleList([TransformerBlock(layer_id, args) for _ in range(self.num_experts)])
        self.gate = GatingMechanism(args.dim, self.num_experts)
        self.layer_id = layer_id

    def forward(self, x: torch.Tensor, start_pos: int, freqs_cis: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        """Forward pass through the MoE Transformer block."""
        gate_values = self.gate(x)  # Shape: [batch_size, seqlen, num_experts]
        expert_outputs = torch.stack([expert(x, start_pos, freqs_cis, mask) for expert in self.experts], dim=-1)  # Shape: [batch_size, seqlen, dim, num_experts]
        weighted_expert_output = torch.einsum('bte,btde->btd', gate_values, expert_outputs)  # Adjusted einsum equation
        return weighted_expert_output

class MoETransformer(nn.Module):
    """Mixture of Experts Transformer model with BitNet-enhanced LLAMA architecture."""
    def __init__(self, params):
        super().__init__()
        self.tok_embeddings = nn.Embedding(params.vocab_size, params.dim)
        self.layers = nn.ModuleList([MoETransformerBlock(layer_id, params) if layer_id % 2 == 0 else TransformerBlock(layer_id, params) for layer_id in range(params.n_layers)])
        self.norm = RMSNorm(params.dim, eps=params.norm_eps)
        self.output = nn.Linear(params.dim, params.vocab_size, bias=False)
        self.freqs_cis = compute_frequency_tensor(params.dim // params.n_heads, params.max_seq_len * 2)

    @torch.no_grad()
    def forward(self, tokens: torch.Tensor, start_pos: int) -> torch.Tensor:
        """Forward pass through the MoE Transformer model."""
        bsz, seqlen = tokens.shape
        h = self.tok_embeddings(tokens)
        self.freqs_cis = self.freqs_cis
        freqs_cis = self.freqs_cis[start_pos : start_pos + seqlen]

        mask = None
        if seqlen > 1:
            mask = torch.full((seqlen, seqlen), float("-inf"), device=tokens.device)
            mask = torch.triu(mask, diagonal=1)
            mask = torch.hstack([torch.zeros((seqlen, start_pos), device=tokens.device), mask]).type_as(h)

        for layer in self.layers:
            h = layer(h, start_pos, freqs_cis, mask)
        h = self.norm(h)
        return self.output(h).float()









import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from typing import Optional, Tuple
from isobits.torch_bitnet import RMSNorm


def compute_frequency_tensor(dim: int, end: int,
                             theta: float = 10000.0) -> torch.Tensor:
    """Calculate a frequency tensor with complex exponentials.

    Args:
        dim (int): Dimension of the frequency tensor.
        end (int): End index for precomputing frequencies.
        theta (float, optional): Scaling factor for frequency computation.
        Defaults to 10000.0.

    Returns:
        torch.Tensor: Precomputed frequency tensor with complex exponentials.
    """
    # Calculate the frequency values
    freqs = 1.0 / (theta ** (torch.arange(0, dim, 2).float() / dim))

    # Create the time index tensor
    t = torch.arange(end)

    # Outer product to get the frequency matrix
    freqs_matrix = torch.outer(t, freqs)

    # Convert frequency matrix to complex values with module 1 and freqs_matrix
    freqs_complex = torch.polar(torch.ones_like(freqs_matrix), freqs_matrix)

    return freqs_complex


def apply_rotary_embeddings(
    xq: torch.Tensor,
    xk: torch.Tensor,
    freqs_complex: torch.Tensor
) -> Tuple[torch.Tensor, torch.Tensor]:
    """Apply rotary embeddings to query and key tensors using the frequency
    tensor.

    Args:
        xq (torch.Tensor): Query tensor to apply rotary embeddings.
        xk (torch.Tensor): Key tensor to apply rotary embeddings.
        freqs_complex (torch.Tensor): Precomputed frequency tensor for complex
        exponentials.

    Returns:
        Tuple[torch.Tensor, torch.Tensor]: Tuple of modified query tensor and
        key tensor with rotary embeddings.
    """
    # Reshape the input tensors to complex numbers
    xq_complex = torch.view_as_complex(xq.float()
                                       .reshape(*xq.shape[:-1], -1, 2))
    xk_complex = torch.view_as_complex(xk.float()
                                       .reshape(*xk.shape[:-1], -1, 2))

    # Reshape the frequency tensor for broadcasting compatibility
    freqs_complex = reshape_freqs_for_broadcast(freqs_complex, xq_complex)

    # Apply the rotary embeddings
    xq_rotary = torch.view_as_real(xq_complex * freqs_complex).flatten(3)
    xk_rotary = torch.view_as_real(xk_complex * freqs_complex).flatten(3)

    return xq_rotary.type_as(xq), xk_rotary.type_as(xk)


def reshape_freqs_for_broadcast(freqs_complex: torch.Tensor,
                                x: torch.Tensor) -> torch.Tensor:
    """Reshape frequency tensor for broadcasting compatibility with the target
    tensor.

    Args:
        freqs_complex (torch.Tensor): Frequency tensor to be reshaped.
        x (torch.Tensor): Target tensor for broadcasting compatibility.

    Returns:
        torch.Tensor: Reshaped frequency tensor.
    """
    # Get the number of dimensions of the target tensor
    ndim = x.ndim

    # Frequency tensor must have same shape as the target tensor
    shape = [1] * ndim
    shape[1] = x.shape[1]  # Match the second dimension
    shape[-1] = x.shape[-1]  # Match the last dimension

    # Reshape the frequency tensor
    return freqs_complex.view(*shape)


def repeat_key_value_tensor(x: torch.Tensor, n_rep: int) -> torch.Tensor:
    """Repeat the key-value tensor along the specified dimension.

    Args:
        x (torch.Tensor): Input tensor to be repeated.
        n_rep (int): Number of repetitions.

    Returns:
        torch.Tensor: Repeated tensor.
    """
    bs, slen, n_kv_heads, head_dim = x.shape
    if n_rep == 1:
        return x
    return (
        x[:, :, :, None, :]
        .expand(bs, slen, n_kv_heads, n_rep, head_dim)
        .reshape(bs, slen, n_kv_heads * n_rep, head_dim)
    )


class MultiheadAttention(nn.Module):
    """Multi-head attention module incorporating rotary embeddings."""
    def __init__(self, args):
        """
        Initialize the MultiheadAttention module.

        Args:
            args: Model configuration parameters.
        """
        super().__init__()
        self.n_kv_heads = args.n_heads if args.n_kv_heads is None else args.n_kv_heads
        self.n_local_heads = args.n_heads
        self.n_local_kv_heads = self.n_kv_heads
        self.n_rep = self.n_local_heads // self.n_local_kv_heads
        self.head_dim = args.dim // args.n_heads

        self.wq = nn.Linear(args.dim, args.n_heads * self.head_dim, bias=False)
        self.wk = nn.Linear(args.dim, self.n_kv_heads * self.head_dim,
                            bias=False)
        self.wv = nn.Linear(args.dim, self.n_kv_heads * self.head_dim,
                            bias=False)
        self.wo = nn.Linear(args.n_heads * self.head_dim, args.dim,
                            bias=False)

        self.cache_k = torch.zeros(args.max_batch_size, args.max_seq_len,
                                   self.n_local_kv_heads, self.head_dim)
        self.cache_v = torch.zeros(args.max_batch_size, args.max_seq_len,
                                   self.n_local_kv_heads, self.head_dim)

    def forward(
        self,
        x: torch.Tensor,
        start_pos: int,
        freqs_complex: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Forward pass of the MultiheadAttention module.

        Args:
            x (torch.Tensor): Input tensor.
            start_pos (int): Starting position for caching.
            freqs_complex (torch.Tensor): Precomputed frequency tensor.
            mask (torch.Tensor, optional): Attention mask tensor.

        Returns:
            torch.Tensor: Output tensor after attention.
        """
        bsz, seqlen, _ = x.shape
        xq, xk, xv = self.wq(x), self.wk(x), self.wv(x)

        xq = xq.view(bsz, seqlen, self.n_local_heads, self.head_dim)
        xk = xk.view(bsz, seqlen, self.n_local_kv_heads, self.head_dim)
        xv = xv.view(bsz, seqlen, self.n_local_kv_heads, self.head_dim)

        xq, xk = apply_rotary_embeddings(xq, xk, freqs_complex=freqs_complex)

        self.cache_k = self.cache_k.to(xq)
        self.cache_v = self.cache_v.to(xq)

        self.cache_k[:bsz, start_pos: start_pos + seqlen] = xk
        self.cache_v[:bsz, start_pos: start_pos + seqlen] = xv

        keys = self.cache_k[:bsz, : start_pos + seqlen]
        values = self.cache_v[:bsz, : start_pos + seqlen]

        keys = repeat_key_value_tensor(keys, self.n_rep)
        values = repeat_key_value_tensor(values, self.n_rep)

        xq = xq.transpose(1, 2)
        keys = keys.transpose(1, 2)
        values = values.transpose(1, 2)
        scores = torch.matmul(xq, keys.transpose(2, 3)) / math.sqrt(self.head_dim)
        if mask is not None:
            scores = scores + mask
        scores = F.softmax(scores.float(), dim=-1).type_as(xq)
        output = torch.matmul(scores, values)
        output = output.transpose(1, 2).contiguous().view(bsz, seqlen, -1)
        return self.wo(output)

class FeedForward(nn.Module):
    """FeedForward module with custom hidden dimension and optional multiplier."""
    def __init__(
        self,
        dim: int,
        hidden_dim: int,
        multiple_of: int,
        ffn_dim_multiplier: Optional[float] = None,
    ):
        super().__init__()
        hidden_dim = int(2 * hidden_dim / 3)
        if ffn_dim_multiplier is not None:
            hidden_dim = int(ffn_dim_multiplier * hidden_dim)
        hidden_dim = multiple_of * ((hidden_dim + multiple_of - 1) // multiple_of)

        self.w1 = nn.Linear(dim, hidden_dim, bias=False)
        self.w2 = nn.Linear(hidden_dim, dim, bias=False)
        self.w3 = nn.Linear(dim, hidden_dim, bias=False)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.w2(F.silu(self.w1(x)) * self.w3(x))


class TransformerBlock(nn.Module):
    """Transformer block consisting of attention and feedforward layers with normalization.

    Args:
        layer_id (int): Identifier for the layer.
        args (ModelArgs): Model configuration parameters.
    """
    def __init__(self, layer_id: int, args):
        super().__init__()
        self.attention = MultiheadAttention(args)
        self.feed_forward = FeedForward(
            dim=args.dim,
            hidden_dim=4 * args.dim,
            multiple_of=args.multiple_of,
            ffn_dim_multiplier=args.ffn_dim_multiplier,
        )
        self.layer_id = layer_id
        self.attention_norm = RMSNorm(args.dim, eps=args.norm_eps)
        self.ffn_norm = RMSNorm(args.dim, eps=args.norm_eps)

    def forward(
        self,
        x: torch.Tensor,
        start_pos: int,
        freqs_complex: torch.Tensor,
        mask: Optional[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Perform a forward pass through the TransformerBlock.

        Args:
            x (torch.Tensor): Input tensor.
            start_pos (int): Starting position for attention caching.
            freqs_complex (torch.Tensor): Precomputed frequency tensor.
            mask (torch.Tensor, optional): Attention mask tensor.

        Returns:
            torch.Tensor: Output tensor after applying attention and feedforward layers.
        """
        h = x + self.attention(self.attention_norm(x), start_pos, freqs_complex, mask)
        return h + self.feed_forward(self.ffn_norm(h))

class Transformer(nn.Module):
    """Transformer model consisting of multiple layers of Transformer blocks.

    Args:
        params (ModelArgs): Model configuration parameters.
    """
    def __init__(self, params):
        super().__init__()
        self.tok_embeddings = nn.Embedding(params.vocab_size, params.dim)
        self.layers = nn.ModuleList([TransformerBlock(layer_id, params) for layer_id in range(params.n_layers)])
        self.norm = RMSNorm(params.dim, eps=params.norm_eps)
        self.output = nn.Linear(params.dim, params.vocab_size, bias=False)
        self.freqs_complex = compute_frequency_tensor(params.dim // params.n_heads, params.max_seq_len * 2)

    @torch.no_grad()
    def forward(self, tokens: torch.Tensor, start_pos: int) -> torch.Tensor:
        """Perform a forward pass through the Transformer model.

        Args:
            tokens (torch.Tensor): Input token indices.
            start_pos (int): Starting position for attention caching.

        Returns:
            torch.Tensor: Output logits after applying the Transformer model.
        """
        bsz, seqlen = tokens.shape
        h = self.tok_embeddings(tokens)
        self.freqs_complex = self.freqs_complex.to(h.device)
        freqs_complex = self.freqs_complex[start_pos : start_pos + seqlen]

        mask = None
        if seqlen > 1:
            mask = torch.full((seqlen, seqlen), float("-inf"), device=tokens.device)
            mask = torch.triu(mask, diagonal=1)
            mask = torch.hstack([torch.zeros((seqlen, start_pos), device=tokens.device), mask]).type_as(h)

        for layer in self.layers:
            h = layer(h, start_pos, freqs_complex, mask)
        h = self.norm(h)
        return self.output(h).float()


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="isoai",  
    version="0.0.1",  
    author="Jazmia Henry",
    author_email="isojaz@isoai.co",
    description="Compress, Trace and Deploy Your Models with Iso!",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/iso-ai/bitmodal",  
    packages=find_packages(),
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=[
        # List your dependencies here
        "torch>=1.7.1",
        # Any other dependencies
    ],
)



"""Vanilla Implementation of BitNet1.58"""
import torch.nn as nn
from isobits.torch_bitnet import TorchBitNet


class BitNet1_5(nn.Module):
    """Implementation of BitNet1.5 as described in the paper:
    'BitNet: Scaling 1-bit Transformers for Large Language Models'.

    Args:
        input_dim (int): Dimension of the input features.
        hidden_dim (int): Dimension of the hidden features.
        output_dim (int): Dimension of the output features.
        norm_dim (int): Dimension for normalization in RMSNorm.
        num_layers (int): Number of layers in the network.
    """
    def __init__(self, input_dim, hidden_dim, output_dim, norm_dim,
                 num_layers):
        super(BitNet1_5, self).__init__()
        self.num_layers = num_layers
        self.layers = nn.ModuleList()

        # Input layer
        self.layers.append(TorchBitNet(input_dim, hidden_dim, norm_dim))

        # Hidden layers
        for _ in range(num_layers - 2):
            self.layers.append(TorchBitNet(hidden_dim, hidden_dim, norm_dim))

        # Output layer
        self.layers.append(TorchBitNet(hidden_dim, output_dim, norm_dim))

    def forward(self, x):
        """
        Forward pass for BitNet1.5.

        Args:
            x (Tensor): Input tensor.

        Returns:
            Tensor: Output tensor after passing through BitNet1.5.
        """
        # Convert quantized output back to float for next layer
        for layer in self.layers:
            x = layer(x).float()
        return x
